#!/usr/bin/env python3

import unittest
from fix_busted_json import repair_json, to_array_of_plain_strings_or_json, first_json, last_json, largest_json, json_matching
import json
import re

class TestParseJson(unittest.TestCase):

    def assert_is_json(self, json_string):
        try:
            json.loads(json_string)
            return True
        except ValueError:
            return False

    def test_to_string_function_exists(self):
        self.assertIsNotNone(repair_json)

    def test_should_return_valid_json_string_from_inspected_output(self):
        input = "{ test: 'test', array: ['test', { test: 'test' }] }"
        result = repair_json(input)
        expected = '{ "test": "test", "array": ["test", { "test": "test" }] }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_return_array_of_strings_and_json_strings(self):
        input = "text before { test: 'test', array: ['test', { test: 'test' }] } text after"
        result = to_array_of_plain_strings_or_json(input)
        self.assertEqual(result[0], 'text before ')
        self.assertEqual(result[1], '{ "test": "test", "array": ["test", { "test": "test" }] }')
        self.assertEqual(result[2], ' text after')
        self.assertTrue(self.assert_is_json(result[1]))

    def test_should_return_first_valid_json_object(self):
        input = "text before { test: 'test', array: ['test', { test: 'test' }] } text { hey: 1 } after"
        result = first_json(input)
        expected = '{ "test": "test", "array": ["test", { "test": "test" }] }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_return_last_valid_json_object(self):
        input = "text before { test: 'test', array: ['test', { test: 'test' }] } text { hey: 1 } after"
        result = last_json(input)
        expected = '{ "hey": 1 }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_return_largest_valid_json_object(self):
        input = "text { gday: 'hi' } before { test: 'test', array: ['test', { test: 'test' }] } text { hey: 1 } after"
        result = largest_json(input)
        expected = '{ "test": "test", "array": ["test", { "test": "test" }] }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_return_json_object_matching_regular_expression(self):
        input = "text { gday: 'hi' } before { test: 'test', array: ['test', { test: 'test' }] } text { hey: 1 } after"
        pattern = re.compile("hey")
        result = json_matching(input, pattern)
        expected = '{ "hey": 1 }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_brace_in_string(self):
        input = 'real json {"value":["closing brace }"]}'
        result = to_array_of_plain_strings_or_json(input)
        self.assertEqual(result[0], 'real json ')
        self.assertEqual(result[1], '{ "value": ["closing brace }"] }')
        self.assertTrue(self.assert_is_json(result[1]))

    def test_should_cope_with_double_quote_in_string(self):
        input = 'real json {"value":["double quote \\"test"]}'
        result = to_array_of_plain_strings_or_json(input)
        self.assertEqual(result[0], 'real json ')
        self.assertEqual(result[1], '{ "value": ["double quote \\"test"] }')
        self.assertTrue(self.assert_is_json(result[1]))

    def test_should_parse_an_array_of_numbers(self):
        input = 'real json { "test": [ 1, 2, 3] }'
        result = to_array_of_plain_strings_or_json(input)
        self.assertEqual(result[0], 'real json ')
        self.assertEqual(result[1], '{ "test": [1, 2, 3] }')
        self.assertTrue(self.assert_is_json(result[1]))

    def test_should_cope_with_empty_object(self):
        input = 'is json {} abc'
        result = to_array_of_plain_strings_or_json(input)
        self.assertEqual(result[0], 'is json ')
        self.assertEqual(result[1], '{  }')
        self.assertEqual(result[2], ' abc')

    def test_should_throw_on_unexpected_end_of_quoted_key_or_string(self):
        with self.assertRaises(IndexError):
            repair_json('{"}')

    def test_should_cope_with_all_kinds_of_whitespace(self):
        input = ' {  \t "test"\t: \t 123 \r \n }'
        result = repair_json(input)
        expected = '{ "test": 123 }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_when_changing_single_quoted_string_to_double_quotes_needs_to_quote_double_quotes(self):
        input = """{ 'abc "': 'test' }"""
        result = repair_json(input)
        expected = '{ "abc \\"": "test" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_when_changing_single_quoted_string_to_double_quotes_needs_to_unquote_single_quotes(self):
        with open('tests/singlequoted.txt', 'r') as file:
            input = file.read()
        result = repair_json(input)
        expected = """{ "abc '": "test'", "key": 123 }"""
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_when_changing_backtick_quoted_string_to_double_quotes_needs_to_fix_quotes(self):
        input = "{ `abc \'\"`: `test\'\"`, 'key': 123}"
        result = repair_json(input)
        expected = """{ "abc '\\"": "test'\\"", "key": 123 }"""
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_when_changing_backticked_quoted_string_to_double_quotes_needs_to_unquote_single_quotes_but_not_double(self):
        with open('tests/backticked.txt', 'r') as file:
            input = file.read()
        result = repair_json(input)
        expected = """{ "abc '\\"`": "test`'\\"", "key": 123 }"""
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_cope_with_trailing_comma_in_key_value_pairs_for_object(self):
        input = '{ "abc": 123, }'
        result = repair_json(input)
        expected = '{ "abc": 123 }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_cope_with_circular_references(self):
        scenario = """{
    abc: <ref *1> {
        abc: 123,
        def: 'test',
        ghi: { jkl: 'test' },
        xyz: { zzz: 123, jj: 'test', abc: [Circular *1] }
    },
    def: <ref *2> {
        zzz: 123,
        jj: 'test',
        abc: <ref *1> {
        abc: 123,
        def: 'test',
        ghi: { jkl: 'test' },
        xyz: [Circular *2]
        }
    }
    }"""
        result = repair_json(scenario)
        self.assertIn('"Circular"', result)
        self.assertTrue(self.assert_is_json(result))

    def test_can_cope_with_stringified_strings(self):
        scenario = '{\n  \"abc\": 123\n  }\n}'
        result = repair_json(scenario)
        expected = '{ "abc": 123 }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_empty_object_is_valid_json(self):
        scenario = '{}'
        result = repair_json(scenario)
        self.assertTrue(self.assert_is_json(result))

    def test_should_play_nice_with_empty_objects(self):
        scenario = '{ "t": [], "a": {} }'
        result = repair_json(scenario)
        expected = '{ "t": [], "a": {  } }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_overly_stringified_objects(self):
        object = {
            'arr"ay': [
                1,
                {
                    'obj"ec\'}t': { "\n\nk\"\'ey": "\"\"''',value" },
                },
                [],
                {},
                True,
                None,
            ],
        }
        scenario = json.dumps(json.dumps(json.dumps(json.dumps(object))))
        result = repair_json(scenario)
        self.assertTrue(self.assert_is_json(result))

    def test_should_concatenate_strings_with_plus(self):
        obj = '{ "abc": "test" + "test2" }'
        result = repair_json(obj)
        expected = '{ "abc": "testtest2" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_concatenate_strings_with_plus_and_preserve_whitespace(self):
        obj = '{ "abc": "test" + "test2" + "test3" }'
        result = repair_json(obj)
        expected = '{ "abc": "testtest2test3" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_concatenate_strings_with_different_quotes(self):
        obj = '{ "abc": \'test\' + `test2` + "test3" + \\"test4\\" }'
        result = repair_json(obj)
        expected = '{ "abc": "testtest2test3test4" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_python_true(self):
        obj = '{ "abc": True }'
        result = repair_json(obj)
        expected = '{ "abc": true }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_python_false(self):
        input_object = '{ "abc": False }'
        result = repair_json(input_object)
        expected = '{ "abc": false }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_change_none_primitive_to_null(self):
        input_object = '{ "abc": None }'
        result = repair_json(input_object)
        expected = '{ "abc": null }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_change_noNe_primitive_to_null(self):
        input_object = "{'intent': {'slots': {'location': noNe}, 'confirmationState': 'None', 'name': 'JobSearch', 'state': 'InProgress'}, 'nluConfidence': 0.8}"
        result = repair_json(input_object)
        expected = '{ "intent": { "slots": { "location": null }, "confirmationState": "None", "name": "JobSearch", "state": "InProgress" }, "nluConfidence": 0.8 }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_treat_space_in_key_name_as_terminator_if_no_in_quotes(self):
        input_object = " { toString } "
        with self.assertRaises(Exception, msg="Expected colon"):
            repair_json(input_object)

    def test_should_support_null_key_name(self):
        input_object = " { [null]: 'test' } "
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_bug_should_support_newline_after_object_before_array_end(self):
        input_object = """{
        savedJobs: [
            {
                external: false
            }
        ]
    }"""
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_support_trailing_comma_in_array_1(self):
        input_object = """{
        savedJobs: [
            {
                external: false
            },
        ]
    }"""
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_support_trailing_comma_in_array_2(self):
        input_object = "{ arr: [1,2,3,]}"
        result = repair_json(input_object)
        expected = '{ "arr": [1, 2, 3] }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_support_trailing_comma_in_array_3(self):
        input_object = "{ arr: [,]}"
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_escaped_double_quotes_used_as_quotes_aka_kibana(self):
        input_object = '{\\"@metadata\\":{\\"beat\\":\\"filebeat\\"}}'
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_escaped_double_quotes_used_as_quotes_inside_strings(self):
        input_object = '{\\"@metadata\\":{\\"message\\":\\"{\\\\"url\\\\": \\\\"hey\\\\"}\\"}}'
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_double_escaped_double_quotes_used_as_quotes_case_1(self):
        input_object = '{ \\\\"test\\\\": \\\\"test1\\\\" }'
        result = repair_json(input_object)
        expected = '{ "test": "test1" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_double_escaped_double_quotes_used_as_quotes_case_2(self):
        input_object = '{\\"@metadata\\":{\\"message\\":\\"{\\\\"url\\\\": \\\\"hey\\\\"}\\"}}'
        result = repair_json(input_object)
        expected = '{ "@metadata": { "message": "{\\"url\\": \\"hey\\"}" } }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_pretty_formatted_sloping_double_quotes_as_output_by_word(self):
        input_object = '{\n"abc": “test”\n}'
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_pretty_formatted_sloping_double_quotes_as_output_by_word_case_2(self):
        input_object = '{“abc”: “def”}'
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_stack_overflow_json(self):
        input_object = '{\nstaus: "Success",\nid: 1,\ndata: [{\'Movie\':\'kung fu panda\',\'% viewed\': 50.5},{\'Movie\':\'kung fu panda 2\',\'% viewed\':1.5}],\nmetadata: {\'filters\':[\'Movie\', \'Percentage Viewed\' ] , \'params\':{\'content\':\'Comedy\', \'type\': \'Movie\'}}\n}'
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_insert_missing_commas_between_key_value_pairs(self):
        input_object = '{\n"abc": "def"\n"ghi": "jkl"\n}'
        result = repair_json(input_object)
        expected = '{ "abc": "def", "ghi": "jkl" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_insert_missing_commas_between_key_value_pairs_case_2(self):
        input_object = '{\n"abc": "def"\n"ghi": "jkl"\n"mno": "pqr"\n}'
        result = repair_json(input_object)
        expected = '{ "abc": "def", "ghi": "jkl", "mno": "pqr" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_insert_missing_commas_between_array_elements(self):
        input_object = '{\n"abc": [\n"def"\n"ghi" 3 true null\n]\n}'
        result = repair_json(input_object)
        expected = '{ "abc": ["def", "ghi", 3, true, null] }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))

    def test_should_error_the_right_way_given_broken_json(self):
        input_object = '{ "test": "bad  { "test": "good" }'
        with self.assertRaises(Exception, msg="Expected colon"):
            repair_json(input_object)

    def test_should_cope_with_a_difficult_scenario(self):
        input_object = '{ \nvalue: true peter: \'fun\' number: 3 somekey: "a string"\narray: [ 2 9234 98234 9 9213840  98213409 98234]\n}'
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))

    def test_should_cope_with_unquoted_single_quote_inside_single_quoted_string_if_an_s_follows(self):
        input_object = "{ 'test': 'test's' }"
        result = repair_json(input_object)
        expected = '{ "test": "test\'s" }'
        self.assertEqual(result, expected)
        self.assertTrue(self.assert_is_json(result))


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python3

import unittest
from fix_busted_json import repair_json, to_array_of_plain_strings_or_json, first_json, last_json, largest_json, json_matching
import json
import re

class TestParseJson(unittest.TestCase):

    def assert_is_json(self, json_string):
        try:
            json.loads(json_string)
            return True
        except ValueError:
            return False

    def test_should_cope_with_escaped_double_quotes_used_as_quotes_inside_strings(self):
        input_object = '{\\"@metadata\\":{\\"message\\":\\"{\\\\"url\\\\": \\\\"hey\\\\"}\\"}}'
        result = repair_json(input_object)
        self.assertTrue(self.assert_is_json(result))


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python3

import json
import re

def log(obj):
    if isinstance(obj, (int, float)):
        print(obj)
    elif isinstance(obj, bool):
        print(obj)
    elif isinstance(obj, dict):
        log_pretty(obj)
        print()
    else:
        log_jsons(obj)

def log_jsons(text):
    array = to_array_of_plain_strings_or_json(text)
    for item in array:
        log_pretty(item)
        if is_json(item):
            log_jsons_in_json(item)
    print()

def log_jsons_in_json(item):
    obj = json.loads(item)

    for key in obj:
        if obj[key] and isinstance(obj[key], dict):
            log_jsons_in_json(json.dumps(obj[key]))
        elif can_parse_json(obj[key]):
            print(f"\nFOUND JSON found in key {key} --->")
            log_jsons(obj[key])

def is_json(text):
    try:
        result = json.loads(text)
        if not result:
            return False
        if not isinstance(result, dict):
            return False
        return True
    except Exception:
        return False

def log_pretty(obj):
    try:
        if isinstance(obj, str):
            json_obj = json.loads(obj)
        else:
            json_obj = obj
        print(json.dumps(json_obj, indent=2))
    except Exception:
        print(obj)

def repair_json(input):
    parse_json = JsonParser(input)
    return parse_json.repair_json()

def to_array_of_plain_strings_or_json(input):
    parse_json = JsonParser(input)
    return parse_json.to_array_of_plain_strings_or_json()

def can_parse_json(input):
    parse_json = JsonParser(input)
    try:
        parse_json.repair_json()
        return True
    except Exception:
        return False

def first_json(input):
    parse_json = JsonParser(input)
    result = parse_json.to_array_of_plain_strings_or_json()
    for item in result:
        if can_parse_json(item):
            return item
    return ""

def last_json(input):
    parse_json = JsonParser(input)
    result = parse_json.to_array_of_plain_strings_or_json()
    for i in range(len(result) - 1, -1, -1):
        if can_parse_json(result[i]):
            return result[i]
    return ""

def largest_json(input):
    parse_json = JsonParser(input)
    result = parse_json.to_array_of_plain_strings_or_json()
    largest = ""
    for item in result:
        if can_parse_json(item) and len(item) > len(largest):
            largest = item
    return largest

def json_matching(input, regex):
    parse_json = JsonParser(input)
    result = parse_json.to_array_of_plain_strings_or_json()
    for item in result:
        if can_parse_json(item) and regex.search(item):
            return item
    return ""

class JsonParser:
    def __init__(self, input):
        self.inspected = self.de_stringify(input)
        self.reset_pointer()
        self.debug = False

    def reset_pointer(self):
        self.position = 0
        self.quoted = ''
        self.checkpoint = 0
        self.checkpoint_quoted = ''
        self.quoted_last_comma_position = None

    def set_checkpoint(self):
        if self.debug:
            print('setCheckpoint', self.position, self.inspected[self.position])
        self.checkpoint = self.position
        self.checkpoint_quoted = self.quoted

    def repair_json(self):
        self.reset_pointer()
        self.eat_object()
        return self.quoted

    def de_stringify(self, string):
        try:
            result = json.loads(string)
            if isinstance(result, str):
                return self.de_stringify(result)
            return string
        except Exception as e:
            return string

    def to_array_of_plain_strings_or_json(self):
        result = []
        self.reset_pointer()
        recovery_position = 0
        while self.position < len(self.inspected):
            self.quoted = ''
            self.eat_plain_text()
            result.append(self.quoted)
            self.quoted = ''
            if self.position >= len(self.inspected):
                break
            if self.inspected[self.position] == '{':
                recovery_position = self.position + 1

                try:
                    self.eat_object()
                except Exception as e:
                    self.quoted += '{'
                    self.position = recovery_position

            result.append(self.quoted)

        return result

    def eat_plain_text(self):
        while self.position < len(self.inspected) and self.inspected[self.position] != '{':
            if self.debug:
                print('eat_plain_text', self.position, self.inspected[self.position])
            self.quoted += self.inspected[self.position]
            self.position += 1

    def eat_object(self):
        if self.debug:
            print('eat_object', self.position, self.inspected[self.position])
        self.eat_whitespace()
        self.eat_open_brace()
        self.eat_key_value_pairs()
        self.eat_whitespace()
        self.eat_close_brace()

    def eat_key_value_pairs(self):
        if self.debug:
            print('eat_key_value_pairs', self.position, self.inspected[self.position])
        while True:
            self.eat_whitespace()
            if self.inspected[self.position] == '}':
                self.remove_trailing_comma_if_present()
                break
            self.quoted_last_comma_position = None
            self.eat_key()
            self.eat_whitespace()
            self.eat_colon()
            self.eat_whitespace()
            self.eat_reference_optional()
            self.eat_whitespace()
            self.eat_value()
            self.quoted_last_comma_position = None
            self.eat_whitespace()

            if self.inspected[self.position] == ',':
                self.eat_comma()
            elif self.inspected[self.position] != '}':
                self.quoted += ', '

    def eat_reference_optional(self):
        if self.inspected[self.position] == '<':
            self.eat_reference()

    def eat_reference(self):
        self.set_checkpoint()
        self.eat_open_angle_bracket()
        self.eat_whitespace()
        self.eat_ref()
        self.eat_whitespace()
        self.eat_asterisk()
        self.eat_whitespace()
        self.eat_number()
        self.eat_whitespace()
        self.eat_close_angle_bracket()

    def eat_open_angle_bracket(self):
        if self.inspected[self.position] != '<':
            raise Exception('Expected open angle bracket')
        self.position += 1

    def eat_ref(self):
        if self.inspected[self.position] != 'r':
            raise Exception('Expected r')
        self.position += 1
        if self.inspected[self.position] != 'e':
            raise Exception('Expected e')
        self.position += 1
        if self.inspected[self.position] != 'f':
            raise Exception('Expected f')
        self.position += 1

    def eat_asterisk(self):
        if self.inspected[self.position] != '*':
            raise Exception('Expected asterisk')
        self.position += 1

    def eat_number(self):
        number_regex = re.compile(r'[0-9]')
        while number_regex.match(self.inspected[self.position]):
            self.position += 1

    def eat_close_angle_bracket(self):
        if self.inspected[self.position] != '>':
            raise Exception('Expected close angle bracket')
        self.position += 1

    def eat_comma_post_value_optional(self):
        if self.inspected[self.position] == ',':
            self.eat_comma()
            return True
        return False

    def eat_whitespace(self):
        whitespace_regex = re.compile(r'\s')
        while whitespace_regex.match(self.inspected[self.position]):
            self.position += 1

    def eat_open_brace(self):
        if self.debug:
            print('eat_open_brace', self.position, self.inspected[self.position])
        if self.inspected[self.position] != '{':
            raise Exception('Expected open brace')
        self.quoted += self.inspected[self.position] + ' '
        self.position += 1

    def eat_close_brace(self):
        if self.debug:
            print('eat_close_brace', self.position, self.inspected[self.position])
        if self.inspected[self.position] != '}':
            raise Exception('Expected close brace')
        self.quoted += ' ' + self.inspected[self.position]
        self.position += 1

    def eat_key(self):
        if self.debug:
            print('eat_key', self.position, self.inspected[self.position])
        if self.get_quote():
            self.eat_quoted_key()
        else:
            self.eat_unquoted_key()

    def get_quote(self):
        if self.inspected[self.position] == "'":
            return "'"
        if self.inspected[self.position] == '"':
            return '"'
        if self.inspected[self.position] == '`':
            return '`'
        if self.inspected[self.position] == '“':
            return '”'
        if self.inspected[self.position] == '\\' and self.inspected[self.position + 1] == '"':
            return '\\"'
        if (
            self.inspected[self.position] == '\\' and
            self.inspected[self.position + 1] == '\\' and
            self.inspected[self.position + 2] == '"'
        ):
            return '\\\\"'
        return False

    def check_quote(self, quote):
        if len(quote) == 1:
            return self.inspected[self.position] == quote
        if len(quote) == 2:
            return (
                self.inspected[self.position] == quote[0] and
                self.inspected[self.position + 1] == quote[1]
            )
        if len(quote) == 3:
            return (
                self.inspected[self.position] == quote[0] and
                self.inspected[self.position + 1] == quote[1] and
                self.inspected[self.position + 2] == quote[2]
            )
        return False

    def eat_quote_additional(self, quote):
        eat_extra = len(quote) - 1
        for _ in range(eat_extra):
            self.position += 1

    def eat_quoted_key(self):
        if self.debug:
            print('eat_quoted_key', self.position, self.inspected[self.position])
        self.set_checkpoint()
        self.throw_if_json_special_character(self.inspected[self.position])
        quote = self.get_quote()
        self.quoted += '"'
        self.position += 1
        self.eat_quote_additional(quote)
        while not self.check_quote(quote):
            self.eat_char_or_escaped_char(quote)
        self.quoted += '"'
        self.position += 1
        self.eat_quote_additional(quote)

    def eat_unquoted_key(self):
        if self.debug:
            print('eat_unquoted_key', self.position, self.inspected[self.position])
        self.set_checkpoint()
        if self.inspected[self.position] == '[':
            return self.eat_null_key()
        self.throw_if_json_special_character(self.inspected[self.position])
        self.quoted += '"'
        while self.inspected[self.position] != ':' and self.inspected[self.position] != ' ':
            if self.get_quote():
                raise Exception('Unexpected quote in unquoted key')
            self.quoted += self.inspected[self.position]
            self.position += 1
        self.quoted += '"'

    def eat_null_key(self):
        if self.debug:
            print('eat_null_key', self.position, self.inspected[self.position])
        if self.inspected[self.position] != '[':
            raise Exception('Expected open bracket')
        self.position += 1
        if self.inspected[self.position].lower() != 'n':
            raise Exception('Expected n')
        self.position += 1
        if self.inspected[self.position].lower() != 'u':
            raise Exception('Expected u')
        self.position += 1
        if self.inspected[self.position].lower() != 'l':
            raise Exception('Expected l')
        self.position += 1
        if self.inspected[self.position].lower() != 'l':
            raise Exception('Expected l')
        self.position += 1
        if self.inspected[self.position] != ']':
            raise Exception('Expected close bracket')
        self.position += 1
        self.quoted += '"null"'

    def throw_if_json_special_character(self, char):
        if char in ['{', '}', '[', ']', ':', ',']:
            raise Exception(f'Unexpected character {char} at position {self.position}')

    def eat_colon(self):
        if self.debug:
            print('eat_colon', self.position, self.inspected[self.position])
        if self.inspected[self.position] != ':':
            raise Exception('Expected colon')
        self.quoted += self.inspected[self.position] + ' '
        self.position += 1

    def eat_value(self):
        if self.debug:
            print('eat_value', self.position, self.inspected[self.position])
        if self.inspected[self.position] == '{':
            self.eat_object()
        elif self.get_quote():
            self.eat_string()
            self.eat_concatenated_strings()
        elif self.inspected[self.position] == '[':
            self.eat_array()
        else:
            self.eat_primitive()

    def eat_string(self):
        if self.debug:
            print('eat_string', self.position, self.inspected[self.position])
        self.set_checkpoint()
        quote = self.get_quote()
        self.quoted += '"'
        self.position += 1
        self.eat_quote_additional(quote)
        while not self.is_end_quote_making_allowance_for_unescaped_single_quote(quote):
            self.eat_char_or_escaped_char(quote)
        self.quoted += '"'
        self.position += 1
        self.eat_quote_additional(quote)

    def eat_concatenated_strings(self):
        if self.debug:
            print('eat_concatenated_strings', self.position, self.inspected[self.position])
        virtual_position = self.eat_virtual_whitespace(self.position + 1)
        if self.inspected[virtual_position] != '+':
            return

        self.position = virtual_position + 1
        self.eat_whitespace()
        self.quoted = self.quoted[:-1]

        quote = self.get_quote()
        self.position += 1
        self.eat_quote_additional(quote)
        while not self.is_end_quote_making_allowance_for_unescaped_single_quote(quote):
            self.eat_char_or_escaped_char(quote)
        self.quoted += '"'
        self.position += 1
        self.eat_quote_additional(quote)

        self.eat_concatenated_strings()

    def is_end_quote_making_allowance_for_unescaped_single_quote(self, quote):
        if quote != "'":
            return self.check_quote(quote)
        try:
            if self.check_quote(quote) and self.inspected[self.position + 1] == 's':
                return False
        except IndexError:
            pass
        return self.check_quote(quote)

    def eat_virtual_whitespace(self, virtual_position):
        if virtual_position >= len(self.inspected):
            return virtual_position - 1
        whitespace_regex = re.compile(r'\s')
        while virtual_position < len(self.inspected) and whitespace_regex.match(self.inspected[virtual_position]):
            virtual_position += 1
        return virtual_position

    def is_double_escaped_double_quote(self):
        if self.position + 2 >= len(self.inspected):
            return False
        return (
            self.inspected[self.position] == '\\' and
            self.inspected[self.position + 1] == '\\' and
            self.inspected[self.position + 2] == '"'
        )

    def eat_char_or_escaped_char(self, quote):
        if self.debug:
            print('eat_char_or_escaped_char', self.position, self.inspected[self.position])
        if self.position >= len(self.inspected):
            raise Exception('Unexpected end of quoted key or string')
        if self.debug:
            print(
                'eatCharOrEscapedChar',
                self.position,
                self.inspected[self.position],
                ' ' + str(ord(self.inspected[self.position])),
            )
        if not self.check_quote(quote) and self.inspected[self.position] == '\\':
            if self.is_double_escaped_double_quote():
                self.position += 1
            if (quote == "'" or quote == '`') and self.inspected[self.position + 1] == quote:
                pass
            else:
                self.quoted += self.inspected[self.position]
            self.position += 1
        if (quote == "'" or quote == '`') and self.inspected[self.position] == '"':
            self.quoted += '\\'
        self.quoted += self.inspected[self.position]
        self.position += 1

    def eat_array(self):
        if self.debug:
            print('eat_array', self.position, self.inspected[self.position])
        if self.inspected[self.position] != '[':
            raise Exception('Expected array')
        self.quoted += self.inspected[self.position]
        self.position += 1

        while True:
            self.eat_whitespace()
            if self.inspected[self.position] == ']':
                self.remove_trailing_comma_if_present()
                break
            self.quoted_last_comma_position = None
            self.eat_circular_optional()
            self.eat_value()
            self.eat_whitespace()

            if self.inspected[self.position] == ',':
                self.eat_comma()
            elif self.inspected[self.position] != ']':
                self.quoted += ', '

        self.eat_close_bracket()

    def remove_trailing_comma_if_present(self):
        if self.quoted_last_comma_position:
            self.quoted = (
                self.quoted[:self.quoted_last_comma_position] +
                self.quoted[self.quoted_last_comma_position + 2:]
            )
        self.quoted_last_comma_position = None

    def eat_circular_optional(self):
        if (
            self.inspected[self.position] == 'C' and
            self.inspected[self.position + 1] == 'i' and
            self.inspected[self.position + 2] == 'r' and
            self.inspected[self.position + 3] == 'c' and
            self.inspected[self.position + 4] == 'u' and
            self.inspected[self.position + 5] == 'l' and
            self.inspected[self.position + 6] == 'a' and
            self.inspected[self.position + 7] == 'r'
        ):
            self.eat_circular()

    def eat_circular(self):
        test_regex = re.compile(r'[Circular *\d]')
        while test_regex.match(self.inspected[self.position]):
            self.position += 1
        self.quoted += '"Circular"'

    def eat_comma(self):
        if self.debug:
            print('eat_comma', self.position, self.inspected[self.position])
        if self.inspected[self.position] != ',':
            raise Exception('Expected comma')
        self.quoted += self.inspected[self.position] + ' '
        self.quoted_last_comma_position = len(self.quoted) - 2
        self.position += 1
        return True

    def eat_close_bracket(self):
        if self.inspected[self.position] != ']':
            raise Exception('Expected close bracket')
        self.quoted += self.inspected[self.position]
        self.position += 1
        return False

    def eat_primitive(self):
        self.set_checkpoint()

        if (
            self.inspected[self.position].lower() == 'f' and
            self.inspected[self.position + 1].lower() == 'a' and
            self.inspected[self.position + 2].lower() == 'l' and
            self.inspected[self.position + 3].lower() == 's' and
            self.inspected[self.position + 4].lower() == 'e'
        ):
            return self.eat_false()

        if (
            self.inspected[self.position].lower() == 'n' and
            self.inspected[self.position + 1].lower() == 'o' and
            self.inspected[self.position + 2].lower() == 'n' and
            self.inspected[self.position + 3].lower() == 'e'
        ):
            return self.eat_none()

        if (
            self.inspected[self.position].lower() == 't' and
            self.inspected[self.position + 1].lower() == 'r' and
            self.inspected[self.position + 2].lower() == 'u' and
            self.inspected[self.position + 3].lower() == 'e'
        ):
            return self.eat_true()

        primitive_regex = re.compile(r'([0-9a-zA-Z-.])')
        while primitive_regex.match(self.inspected[self.position]):
            self.quoted += self.inspected[self.position]
            self.position += 1

    def eat_false(self):
        self.quoted += 'false'
        self.position += 5

    def eat_none(self):
        self.quoted += 'null'
        self.position += 4

    def eat_true(self):
        self.quoted += 'true'
        self.position += 4
